#include <thrust/host_vector.h>
#include <thrust/device_vector.h>
#include <thrust/generate.h>
#include <thrust/sort.h>
#include <thrust/copy.h>

#include <algorithm>
#include <cstdlib>
#include <iostream>
#include <fstream>
#include <cassert>
#include <vector>

#include <time.h>
#include <sys/timeb.h>
#include <stdint.h>

#include <tbb/parallel_sort.h>
#include <tbb/task_scheduler_init.h>

#include <boost/smart_ptr.hpp> 

#include <hype.hpp>

#include <query_processing/operator.hpp>
#include <query_processing/processing_device.hpp>

using namespace std;

	typedef int ElementType;
	typedef vector<ElementType> Vec;
	typedef boost::shared_ptr<Vec> VecPtr;

	enum SchedulingConfiguration{CPU_ONLY,GPU_ONLY,HYBRID};


        uint64_t getTimestamp()
        {
  		   timespec ts;
   		clock_gettime(CLOCK_REALTIME, &ts);
         return (uint64_t)ts.tv_sec * 1000000000LL + (uint64_t)ts.tv_nsec;
        }

struct Random_Number_Generator{

	Random_Number_Generator(unsigned int max_value_size) : max_value_size_(max_value_size){}

	ElementType operator() (){
		return (ElementType) rand()%max_value_size_;
	}
	private:
		unsigned int max_value_size_;
};



void CPU_Sort(VecPtr dataset){
	assert(dataset!=NULL);
	std::sort(dataset->begin(),dataset->end());
}

void CPU_Sort_Parallel(VecPtr dataset){
	assert(dataset!=NULL);
	tbb::task_scheduler_init init(4);
	tbb::parallel_sort(dataset->begin(),dataset->end());
}

void GPU_Sort(VecPtr dataset){
	assert(dataset!=NULL);
	// transfer data to the device
  thrust::device_vector<ElementType> d_vec (dataset->begin(),dataset->end());

  // sort data on the device (846M keys per second on GeForce GTX 480)
  thrust::sort(d_vec.begin(), d_vec.end());

  // transfer data back to host
  thrust::copy(d_vec.begin(), d_vec.end(), dataset->begin());
}

	class CPU_Serial_Sort_Operator : public hype::queryprocessing::Operator{
		public:
		CPU_Serial_Sort_Operator(const hype::SchedulingDecision& sd, VecPtr input_data) : Operator(sd), input_data_(input_data){

		}

		virtual bool execute(){
			CPU_Sort(input_data_);
			return true;
		}

		VecPtr input_data_;		
	};

	class CPU_Parallel_Sort_Operator : public hype::queryprocessing::Operator{
		public:
		CPU_Parallel_Sort_Operator(const hype::SchedulingDecision& sd, VecPtr input_data) : Operator(sd), input_data_(input_data){

		}

		virtual bool execute(){
			CPU_Sort_Parallel(input_data_);
			return true;
		}

		VecPtr input_data_;		
	};

	class GPU_Sort_Operator : public hype::queryprocessing::Operator{
		public:
		GPU_Sort_Operator(const hype::SchedulingDecision& sd, VecPtr input_data) : Operator(sd), input_data_(input_data){

		}

		virtual bool execute(){
			std::cout << "[GPU_Sort_Operator] Exec GPU Sort" << std::endl;
			GPU_Sort(input_data_);
			return true;
		}

		VecPtr input_data_;		
	};


VecPtr generate_dataset(unsigned int size_in_number_of_elements){
	VecPtr data(new Vec());
	for(unsigned int i=0;i<size_in_number_of_elements;i++){
		ElementType e = (ElementType) rand();
		data->push_back(e);
	}
	assert(data!=NULL);
	std::cout << "created new data set: " << data.get() << " of size: " << data->size() << std::endl;
	return data;
}

vector<VecPtr> generate_random_datasets(unsigned int max_size_in_number_of_elements, unsigned int number_of_datasets){
	vector<VecPtr> datasets;
		for(unsigned int i=0;i<number_of_datasets;i++){
			VecPtr vec_ptr = generate_dataset((unsigned int) (rand()%max_size_in_number_of_elements) );
			assert(vec_ptr!=NULL);
			datasets.push_back(vec_ptr);
		}
	return datasets;
}

int main(int argc, char* argv[]){


//	tbb::task_scheduler_init init(8);
	
	srand(0);
	//tbb::task_scheduler_init (2);
//	tbb::task_scheduler_init init(1);
//	


	cout << "TBB use " << tbb::task_scheduler_init::default_num_threads() << " number of threads as a default" << endl;
	const unsigned int MAX_DATASET_SIZE=(10*1000*1000)/sizeof(int); //1000000;
	const unsigned int NUMBER_OF_DATASETS=3; //100;
	const unsigned int NUMBER_OF_SORT_OPERATIONS_IN_WORKLOAD=1; //3; //1000;

	//SchedulingConfiguration sched_config=HYBRID; //CPU_ONLY,GPU_ONLY,HYBRID
//	SchedulingConfiguration sched_config=GPU_ONLY;
	SchedulingConfiguration sched_config=CPU_ONLY;

	cout << "Generating Data sets..." << endl;

	vector<VecPtr> datasets=generate_random_datasets(MAX_DATASET_SIZE, NUMBER_OF_DATASETS);
	vector<unsigned int> query_indeces(NUMBER_OF_SORT_OPERATIONS_IN_WORKLOAD);

	std::generate(query_indeces.begin(), query_indeces.end(), Random_Number_Generator(NUMBER_OF_DATASETS));

	std::copy(query_indeces.begin(), query_indeces.end(), std::ostream_iterator<unsigned int>(std::cout, "\n"));

	//setup STEMOD
	//hype::Scheduler::instance().addAlgorithm("SORT","CPU_Algorithm_serial","Least Squares 1D","Periodic Recomputation");
	hype::Scheduler::instance().addAlgorithm("SORT","CPU_Algorithm_parallel", hype::CPU, "Least Squares 1D", "Periodic Recomputation");
	hype::Scheduler::instance().addAlgorithm("SORT","GPU_Algorithm", hype::GPU, "Least Squares 1D", "Periodic Recomputation");

	hype::queryprocessing::ProcessingDevice cpu;
	hype::queryprocessing::ProcessingDevice gpu;
		
	cpu.start();
	gpu.start();

	//boost::this_thread::sleep( boost::posix_time::seconds(30) );

	cout << "Starting Benchmark..." << endl;

	uint64_t begin_benchmark_timestamp = getTimestamp();

	for(unsigned int i=0;i<NUMBER_OF_SORT_OPERATIONS_IN_WORKLOAD;i++){
		unsigned int index = query_indeces[i];
		VecPtr dataset = datasets[index];
	
		assert(NUMBER_OF_SORT_OPERATIONS_IN_WORKLOAD==query_indeces.size());
		assert(index<NUMBER_OF_DATASETS); //NUMBER_OF_SORT_OPERATIONS_IN_WORKLOAD);	
	
		hype::Tuple t;
		t.push_back(dataset->size());
		hype::SchedulingDecision sched_dec_local("",hype::core::EstimatedTime(0),t);
	

		//cout << "RUN: " << i << "/" << NUMBER_OF_SORT_OPERATIONS_IN_WORKLOAD << endl;
		
		if(sched_config==HYBRID){ //CPU_ONLY,GPU_ONLY,HYBRID)
			hype::SchedulingDecision sched_dec = hype::Scheduler::instance().getOptimalAlgorithmName("SORT",t);

		cout << "Decision: " << sched_dec.getNameofChoosenAlgorithm() << endl;

//		if(sched_dec.getNameofChoosenAlgorithm()=="CPU_Algorithm_serial"){
//			hype::AlgorithmMeasurement alg_measure(sched_dec);
//			CPU_Sort(dataset);
//			alg_measure.afterAlgorithmExecution(); 
//		}else if(sched_dec.getNameofChoosenAlgorithm()=="CPU_Algorithm_parallel"){
//			hype::AlgorithmMeasurement alg_measure(sched_dec);
//			CPU_Sort_Parallel(dataset);
//			alg_measure.afterAlgorithmExecution(); 
//		}else if(sched_dec.getNameofChoosenAlgorithm()=="GPU_Algorithm"){
//			hype::AlgorithmMeasurement alg_measure(sched_dec);
//			GPU_Sort(dataset);
//			alg_measure.afterAlgorithmExecution(); 
//		}

		if(sched_dec.getNameofChoosenAlgorithm()=="CPU_Algorithm_serial"){
			cpu.addOperator( boost::shared_ptr<CPU_Serial_Sort_Operator>( new CPU_Serial_Sort_Operator(sched_dec, dataset) ) );
//			hype::AlgorithmMeasurement alg_measure(sched_dec);
//			CPU_Sort(dataset);
//			alg_measure.afterAlgorithmExecution(); 
		}else if(sched_dec.getNameofChoosenAlgorithm()=="CPU_Algorithm_parallel"){
			cpu.addOperator( boost::shared_ptr<CPU_Parallel_Sort_Operator>( new CPU_Parallel_Sort_Operator(sched_dec, dataset) ) );
//			hype::AlgorithmMeasurement alg_measure(sched_dec);
//			CPU_Sort_Parallel(dataset);
//			alg_measure.afterAlgorithmExecution(); 
		}else if(sched_dec.getNameofChoosenAlgorithm()=="GPU_Algorithm"){
			gpu.addOperator( boost::shared_ptr<GPU_Sort_Operator>( new GPU_Sort_Operator(sched_dec, dataset) ) );
//			hype::AlgorithmMeasurement alg_measure(sched_dec);
//			GPU_Sort(dataset);
//			alg_measure.afterAlgorithmExecution(); 
		}

		}else if(sched_config==CPU_ONLY){
			//CPU_Sort_Parallel(dataset);
			std::cout << "Assigning Operator to CPU... " << std::endl;
			cpu.addOperator( boost::shared_ptr<CPU_Parallel_Sort_Operator>( new CPU_Parallel_Sort_Operator(sched_dec_local, dataset) ) );
		}else if(sched_config==GPU_ONLY){
			//GPU_Sort(dataset);
			std::cout << "Assigning Operator to GPU... " << std::endl;
			gpu.addOperator( boost::shared_ptr<GPU_Sort_Operator>( new GPU_Sort_Operator(sched_dec_local, dataset) ) );
		}

	}

//	boost::this_thread::sleep( boost::posix_time::seconds(3) );

//	cpu.stop();
//	gpu.stop();

	while(!cpu.isIdle() || !gpu.isIdle()){

	}

	cpu.stop();
	gpu.stop();

	uint64_t end_benchmark_timestamp = getTimestamp();

	assert(end_benchmark_timestamp>=begin_benchmark_timestamp);
	
	cout << "Time for Workload: " <<  end_benchmark_timestamp-begin_benchmark_timestamp << "ns (" 
		  << double(end_benchmark_timestamp-begin_benchmark_timestamp)/(1000*1000*1000) << "s)" << endl;

 return 0;
}

