//default includes
#include <limits>

#include <config/global_definitions.hpp>

#include <core/operation.hpp>

#include <plugins/optimization_criterias/response_time_advanced.hpp>

#include <query_processing/processing_device.hpp>

#ifdef DUMP_ESTIMATIONS
	#include <fstream>
#endif

//#define TIMESTAMP_BASED_LOAD_ADAPTION
//#define PROBABILITY_BASED_LOAD_ADAPTION
//#define LOAD_MODIFICATOR_BASED_LOAD_ADAPTION

using namespace std;

namespace hype{
	namespace core{

	WaitingTimeAwareResponseTime::WaitingTimeAwareResponseTime(const std::string& name_of_operation) : OptimizationCriterion(name_of_operation,std::string("Response Time")) {
		//OptimizationCriterionFactorySingleton::Instance().Register("Response Time",&WaitingTimeAwareResponseTime::create);
	}

	const SchedulingDecision WaitingTimeAwareResponseTime::getOptimalAlgorithm_internal(const Tuple& input_values, Operation& op, DeviceConstraint dev_constr){
		//idea: choose algorithm with minimal finishing time (including expected waiting time on device!)
		
		assert(dev_constr==hype::ANY_DEVICE);

		std::vector<AlgorithmPtr> alg_ptrs = op.getAlgorithms();
		
		std::vector<double> estimated_finishing_times_(alg_ptrs.size());
		
		std::map<hype::ComputeDevice,double> map_compute_device_to_estimated_waiting_time;
		
		map_compute_device_to_estimated_waiting_time.insert(
			std::make_pair<hype::ComputeDevice,double>(hype::CPU,hype::queryprocessing::getProcessingDevice(hype::CPU).getEstimatedTimeUntilIdle())
		);
		
		map_compute_device_to_estimated_waiting_time.insert(
			std::make_pair<hype::ComputeDevice,double>(hype::GPU,hype::queryprocessing::getProcessingDevice(hype::GPU).getEstimatedTimeUntilIdle())
		);
		
		for(unsigned int i=0;i<estimated_finishing_times_.size();++i){
			//get estimated waiting time, until the operator can start execution
			estimated_finishing_times_[i]=map_compute_device_to_estimated_waiting_time[alg_ptrs[i]->getComputeDevice()];
			//debug output
			if(!quiet && verbose) 
				cout << "estimated_finishing_time of Algorithm " << alg_ptrs[i]->getName() << ": " << estimated_finishing_times_[i] << "ns" << endl;
			//add the estiamted time, the operator itself needs for execution
			estimated_finishing_times_[i]+=alg_ptrs[i]->getEstimatedExecutionTime(input_values).getTimeinNanoseconds();
			//debug output
			if(!quiet && verbose) 
				cout << "estimated_finishing_time of Algorithm " <<alg_ptrs[i]->getName() << " (including Algorithm execution time): " << estimated_finishing_times_[i] << "ns" << endl;
		}
		
		AlgorithmPtr optimal_algorithm_ptr;
		double min_time=std::numeric_limits<double>::max();
		for(unsigned int i=0;i<estimated_finishing_times_.size();++i){
			if(min_time>estimated_finishing_times_[i]){
				min_time=estimated_finishing_times_[i];
				optimal_algorithm_ptr=alg_ptrs[i];
			}
		}
		assert(optimal_algorithm_ptr!=NULL);

		if(!quiet && verbose) 
			cout << "Choosing " << optimal_algorithm_ptr->getName() << " for operation " << op.getName() << endl;
		return SchedulingDecision(*optimal_algorithm_ptr,optimal_algorithm_ptr->getEstimatedExecutionTime(input_values),input_values);

/*
		std::map<double,std::string> map_execution_times_to_algorithm_name = op.getEstimatedExecutionTimesforAlgorithms(input_values);
		if(map_execution_times_to_algorithm_name.empty()){
			 std::cout << "FATAL ERROR! no algorithm to choose from!!!" << std::endl;
			 std::cout << "File: " <<  __FILE__ << " Line: " << __LINE__ << std::endl;
			 exit(-1);
		}
		
		std::map<double,std::string>::iterator it;
		double min_time=std::numeric_limits<double>::max(); 
		for(it=map_execution_times_to_algorithm_name.begin();it!=map_execution_times_to_algorithm_name.end();++it){
			if(!quiet && verbose) cout << "Algorithm: '" << it->second << "'	Estimated Execution Time: " << it->first << endl;
			if(it->first<min_time){
				min_time=it->first;
			}
			#ifdef DUMP_ESTIMATIONS
				std::string path = "output/";
				path+=op.getName()+"/";
				path+=it->second+".estimations";
				fstream file(path.c_str(),fstream::out | fstream::app);
				for(unsigned int i=0;i<input_values.size();++i){
				file << input_values[i] << ",";
				}
				file << it->first << "," << it->second << endl;
				file.close();
			#endif
		}

#ifdef PROBABILITY_BASED_LOAD_ADAPTION
		//FEATURE: probability based load adaption (pick optimal algorithm in X% of all cases and pick another one in the other cases) %%%%%%%%%%%%%%%%%%%%%%%%%%%
//		if(rand()%5000==0){
//			for(it=map_execution_times_to_algorithm_name.begin();it!=map_execution_times_to_algorithm_name.end();it++){
//				if(!quiet && verbose) cout << "Algorithm: '" << it->second << "'	Estimated Execution Time: " << it->first << endl;
//				cout << "Recoice: " << << endl;
//				if(!(it->first<min_time)){
//					min_time=it->first;
//				}
//			}		
//		}
#endif
//		
		std::string name_of_algorithm = map_execution_times_to_algorithm_name[min_time];
		if(!quiet && verbose) cout << "Choosing " << name_of_algorithm << " for operation " << op.getName() << endl;
		return SchedulingDecision(name_of_algorithm,EstimatedTime(min_time),input_values);
		 */
	}

	}; //end namespace core
}; //end namespace stemod


