
#include <core/scheduler.hpp>
#include <plugins/pluginloader.hpp>
#include <stdlib.h>

#include <boost/thread.hpp>

using namespace std;

namespace hype{
	namespace core{

		boost::mutex global_mutex;
		
		Scheduler::Scheduler() : map_operationname_to_operation_(), map_statisticalmethodname_to_statisticalmethod_(){
			//is called once, because Scheduler is a Singelton
			PluginLoader::loadPlugins();
		}

		bool Scheduler::addAlgorithm(const std::string& name_of_operation, 
										const std::string& name_of_algorithm, 
										ComputeDevice comp_dev,
										const std::string& name_of_statistical_method, 
										const std::string& name_of_recomputation_strategy){
											
			boost::lock_guard<boost::mutex> lock(global_mutex);
			
			std::map<std::string,std::tr1::shared_ptr<Operation> >::iterator it;

			it=map_operationname_to_operation_.find(name_of_operation);

			std::tr1::shared_ptr<Operation> op;

			if(it==map_operationname_to_operation_.end()){ //operation does not exist
				op = std::tr1::shared_ptr<Operation>(new Operation(name_of_operation));
				map_operationname_to_operation_[name_of_operation]=op;
			}else{
				op=it->second; //std::tr1::shared_ptr<Operation> (it->second);
			}

			return op->addAlgorithm(name_of_algorithm, 
											comp_dev,
										   name_of_statistical_method, 
										   name_of_recomputation_strategy);


		 //return true;
		}

		bool Scheduler::setOptimizationCriterion(const std::string& name_of_operation,
														const std::string& name_of_optimization_criterion){
															
			boost::lock_guard<boost::mutex> lock(global_mutex);
			
//			map_operationname_to_operation_.find
//			StatisticalMethodMap::iterator it = map_statisticalmethodname_to_statisticalmethod_.find(name_of_operation);
			
			std::tr1::shared_ptr<OptimizationCriterion> opt_crit = getNewOptimizationCriterionbyName(name_of_optimization_criterion);
			if(opt_crit){
				//StatisticalMethodMap::iterator it = map_statisticalmethodname_to_statisticalmethod_.find(name_of_operation);
				
				MapNameToOperation::iterator it = map_operationname_to_operation_.find(name_of_operation);
				if(it==map_operationname_to_operation_.end()){
					cout << "Operation not found! " << name_of_operation << endl;
					return false;
				}
				return (*it).second->setNewOptimizationCriterion(name_of_optimization_criterion);
				//map_operationname_to_operation_[name_of_operation]=opt_crit;
				//return true;
			}
		   return false;
		}

		/* not part of intrface!!! -> no thread safety implemented!!!*/
		const AlgorithmPtr Scheduler::getAlgorithm(const std::string& name_of_algorithm){
			
			//boost::lock_guard<boost::mutex> lock(global_mutex);
			
			MapNameToOperation::iterator it;
			for(it= map_operationname_to_operation_.begin(); it !=map_operationname_to_operation_.end(); it++){
				if(it->second->hasAlgorithm(name_of_algorithm))
					 return it->second->getAlgorithm(name_of_algorithm);	
			}
			return AlgorithmPtr(); //NULL Pointer if algorithm is not found
		}


		bool Scheduler::setStatisticalMethod(const std::string& name_of_algorithm,
												  const std::string& name_of_statistical_method){
													  
			boost::lock_guard<boost::mutex> lock(global_mutex);
			
			AlgorithmPtr alg_ptr = this->getAlgorithm(name_of_algorithm);
			if(!alg_ptr){
				cout << "Error in Scheduler::setStatisticalMethod(): Algorithm " << name_of_algorithm << " not found!" << endl;
				return false;
			}
			cout << "Found Algorithm: " << alg_ptr->getName() << endl;
			StatisticalMethodPtr stat_meth_ptr = getNewStatisticalMethodbyName(name_of_statistical_method);
			if(!stat_meth_ptr) return false;
			cout << "created statistical method: " << name_of_statistical_method << endl;		
			return alg_ptr->setStatisticalMethod(stat_meth_ptr);
		}
		
		bool Scheduler::setRecomputationHeuristic(const std::string& name_of_algorithm,
												  		 const std::string& name_of_recomputation_strategy){
	
			boost::lock_guard<boost::mutex> lock(global_mutex);
			
			AlgorithmPtr alg_ptr = this->getAlgorithm(name_of_algorithm);
			if(!alg_ptr){
				cout << "Error in Scheduler::setRecomputationHeuristic(): Algorithm " << name_of_algorithm << " not found!" << endl;
				return false;
			}
			std::tr1::shared_ptr<RecomputationHeuristic> recomp_heuristic = getNewRecomputationHeuristicbyName(name_of_recomputation_strategy);
			if(!recomp_heuristic) return false;
			return alg_ptr->setRecomputationHeuristic(recomp_heuristic);
		}

		const SchedulingDecision Scheduler::getOptimalAlgorithmName(const std::string& name_of_operation, const Tuple& input_values, DeviceConstraint dev_constr){

			boost::lock_guard<boost::mutex> lock(global_mutex);
			
			std::map<std::string,std::tr1::shared_ptr<Operation> >::iterator it;

			it=map_operationname_to_operation_.find(name_of_operation);

			std::tr1::shared_ptr<Operation> op;

			if(it==map_operationname_to_operation_.end()){ //operation does not exist
				std::cout << "FATAL ERROR: Operation " << name_of_operation  <<  " does not exist!!!" << std::endl;
				exit(-1);
			}//else{
				
				op=it->second; //std::tr1::shared_ptr<Operation> (it->second);
				if(!op){
					std::cout << "FATAL ERROR: Operation " << name_of_operation  <<  ": NULL Pointer!" << std::endl;
					exit(-1);
				}
				return op->getOptimalAlgorithm(input_values, dev_constr);
			//}

		   //return SchedulingDecision(std::string(),EstimatedTime(-1.0),Tuple());
		}
		bool Scheduler::addObservation(const std::string& name_of_algorithm, const MeasurementPair& mp){
			
			boost::lock_guard<boost::mutex> lock(global_mutex);
			
			MapNameToOperation::iterator it;
			for(it= map_operationname_to_operation_.begin(); it !=map_operationname_to_operation_.end(); it++){
				if(it->second->hasAlgorithm(name_of_algorithm))
					 return it->second->addObservation(name_of_algorithm,mp);			
			}
			return false;
		}


	};//end namespace core
}; //end namespace stemod


